@@ -71,23 +71,26 @@ class DataFileSystem(AbstractFileSystem):  # pylint:disable=abstract-method
         return fs.open(fspath, mode=mode, encoding=encoding)
 
     def ls(self, path, detail=True, **kwargs):
-        info = self.info(path)
-        if info["type"] != "directory":
-            return [info] if detail else [path]
-
         root_key = self._get_key(path)
-        try:
-            entries = [
-                self.path.join(path, name) if path else name
-                for name in self.index.ls(prefix=root_key)
-            ]
-        except KeyError as exc:
-            raise FileNotFoundError from exc
+        info = self.index.info(root_key)
+        if info["type"] != "directory":
+            if detail:
+                info["name"] = path
+                return [info]
+            else:
+                return [path]
 
         if not detail:
-            return entries
+            return [
+                self.path.join(path, key[-1])
+                for key in self.index.ls(root_key, detail=False)
+            ]
 
-        return [self.info(epath) for epath in entries]
+        entries = []
+        for key, info in self.index.ls(root_key, detail=True):
+            info["name"] = self.path.join(path, key[-1])
+            entries.append(info)
+        return entries
 
     def isdvc(self, path, recursive=False):
         try:
@@ -102,39 +105,10 @@ class DataFileSystem(AbstractFileSystem):  # pylint:disable=abstract-method
         return self.index.has_node(key)
 
     def info(self, path, **kwargs):
-        from dvc_data.index import ShortKeyError
-
         key = self._get_key(path)
-
-        try:
-            entry = self.index[key]
-            isdir = entry.hash_info and entry.hash_info.isdir
-            return {
-                "type": "directory" if isdir else "file",
-                "name": path,
-                "size": entry.meta.size if entry.meta else 0,
-                "isexec": entry.meta.isexec if entry.meta else False,
-                "version_id": entry.meta.version_id if entry.meta else None,
-                "isdvc": True,
-                "isout": True,
-                "obj": entry.obj,
-                "entry": entry,
-                entry.hash_info.name: entry.hash_info.value,
-            }
-        except ShortKeyError:
-            return {
-                "type": "directory",
-                "name": path,
-                "size": 0,
-                "isexec": False,
-                "version_id": None,
-                "isdvc": bool(self.index.longest_prefix(key)),
-                "isout": False,
-                "obj": None,
-                "entry": None,
-            }
-        except KeyError as exc:
-            raise FileNotFoundError from exc
+        info = self.index.info(key)
+        info["name"] = path
+        return info
 
     def get_file(  # pylint: disable=arguments-differ
         self, rpath, lpath, callback=DEFAULT_CALLBACK, **kwargs
