@@ -4,6 +4,7 @@ from attrs import define
 
 if TYPE_CHECKING:
     from .hashfile.meta import Meta
+    from .hashfile.hash_info import HashInfo
     from .index import BaseDataIndex, DataIndexKey
 
 from .index import DataIndexEntry
@@ -39,6 +40,68 @@ class Change:
         return self.typ != UNCHANGED
 
 
+def _diff_meta(
+    old: Optional["Meta"],
+    new: Optional["Meta"],
+    *,
+    cmp_key: Optional[Callable[["Meta"], Any]] = None,
+):
+    if (cmp_key is None or old is None or new is None) and old != new:
+        return MODIFY
+
+    if cmp_key is not None and cmp_key(old) != cmp_key(new):
+        return MODIFY
+
+    return UNCHANGED
+
+
+def _diff_hash_info(
+    old: Optional["HashInfo"],
+    new: Optional["HashInfo"],
+):
+    if not old and new:
+        return ADD
+
+    if old and not new:
+        return DELETE
+
+    if old and new and old != new:
+        return MODIFY
+
+    return UNCHANGED
+
+
+def _diff_entry(
+    old: Optional["DataIndexEntry"],
+    new: Optional["DataIndexEntry"],
+    *,
+    hash_only: Optional[bool] = False,
+    meta_only: Optional[bool] = False,
+    meta_cmp_key: Optional[Callable[["Meta"], Any]] = None,
+):
+    old_hi = old.hash_info if old else None
+    new_hi = new.hash_info if new else None
+    old_meta = old.meta if old else None
+    new_meta = new.meta if new else None
+
+    meta_diff = _diff_meta(old_meta, new_meta, cmp_key=meta_cmp_key)
+    hi_diff = _diff_hash_info(old_hi, new_hi)
+
+    if meta_only:
+        return meta_diff
+
+    if hash_only:
+        return hi_diff
+
+    if meta_diff != UNCHANGED:
+        return MODIFY
+
+    if hi_diff != UNCHANGED:
+        return MODIFY
+
+    return UNCHANGED
+
+
 def _diff(
     old: Optional["BaseDataIndex"],
     new: Optional["BaseDataIndex"],
@@ -47,35 +110,32 @@ def _diff(
     hash_only: Optional[bool] = False,
     meta_only: Optional[bool] = False,
     meta_cmp_key: Optional[Callable[["Meta"], Any]] = None,
+    shallow: Optional[bool] = False,
 ):
-    old_keys = {key for key, _ in old.iteritems()} if old else set()
-    new_keys = {key for key, _ in new.iteritems()} if new else set()
+    old_keys = (
+        {key for key, _ in old.iteritems(shallow=shallow)} if old else set()
+    )
+    new_keys = (
+        {key for key, _ in new.iteritems(shallow=shallow)} if new else set()
+    )
 
     for key in old_keys | new_keys:
         old_entry = old.get(key) if old is not None else None
         new_entry = new.get(key) if new is not None else None
-        old_hi = old_entry.hash_info if old_entry else None
-        new_hi = new_entry.hash_info if new_entry else None
-        old_meta = old_entry.meta if old_entry else None
-        new_meta = new_entry.meta if new_entry else None
 
         typ = UNCHANGED
         if old_entry and not new_entry:
             typ = DELETE
         elif not old_entry and new_entry:
             typ = ADD
-        elif not meta_only and (old_hi and new_hi):
-            if old_hi != new_hi:
-                typ = MODIFY
-        elif not hash_only:
-            if (
-                meta_cmp_key is None or old_meta is None or new_meta is None
-            ) and old_meta != new_meta:
-                typ = MODIFY
-            elif meta_cmp_key is not None and meta_cmp_key(
-                old_meta
-            ) != meta_cmp_key(new_meta):
-                typ = MODIFY
+        else:
+            typ = _diff_entry(
+                old_entry,
+                new_entry,
+                hash_only=hash_only,
+                meta_only=meta_only,
+                meta_cmp_key=meta_cmp_key,
+            )
 
         if typ == UNCHANGED and not with_unchanged:
             continue
@@ -138,6 +198,7 @@ def diff(
     hash_only: Optional[bool] = False,
     meta_only: Optional[bool] = False,
     meta_cmp_key: Optional[Callable[["Meta"], Any]] = None,
+    shallow: Optional[bool] = False,
 ):
     changes = _diff(
         old,
@@ -146,6 +207,7 @@ def diff(
         hash_only=hash_only,
         meta_only=meta_only,
         meta_cmp_key=meta_cmp_key,
+        shallow=shallow,
     )
 
     if with_renames and old is not None and new is not None:
